/***********************************************************************\ 
*                                                                       * 
*   File: calendric_system/tadt/tadt_arithmetic.c
*                                                                       * 
*   The MultiCal System is free software in the public domain; you can  * 
*   redistribute it and/or modify it as you wish. We ask that you       * 
*   retain credits referencing the University of Arizona and that you   * 
*   identify any changes you make.                                      * 
*                                                                       * 
*   Report problems to multical@cs.arizona.edu                          * 
*   Direct all inquiries to:    The MultiCal Project                    * 
*                               Department of Computer Science          * 
*                               University of Arizona                   * 
*                               Tucson, AZ 85721                        * 
*                               U.S.A.                                  * 
*                                                                       * 
\***********************************************************************/ 
#include "system_wide.h"
#include "tadt.h"

extern resolution_type combined_flags[MAX_RESOLUTIONS_OF_EVENTS][MAX_RESOLUTIONS_OF_EVENTS];

static error_type rc = tadt_ok;
static int32 temp32;
static int64 temp64;
static int96 temp96;

/*  
*  Routine:  tadt_neg_fs
*
*  Description: Determine the sign opposite fixed span which is otherwise the
*		same as the input span.  Has no effect on a zero length
*               span.
*
*  Arguments: 	span -- (IN) : span to be converted
*               result -- (OUT) : result span 
*
*  Return Value: Error code
*
*  Errors: out_of_range (Max negative span has no positive counterpart)
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_neg_fs(span_type span, span_type result) {
  
  result->upack.modified = TRUE;
  switch (span->upack.ts_type) {
    case upack32:
      result->upack.ts_type = upack32;
      LOAD32(-1,temp32);
      MULT32(temp32,span->upack32.granules,result->upack32.granules); 
      return tadt_ok;
      break;

    case upack64: 
      result->upack.ts_type = upack64;
      LOAD64(-1,temp64); 
      MULT64(temp64,span->upack64.granules,result->upack64.granules); 
      return tadt_ok;
      break;

    case upack96: 
      LOAD96(-1,temp96);
      MULT96(temp96,span->upack96.granules,result->upack96.granules); 
      return tadt_ok;
      break;

    case special:
      result->special.ts_type = special;
      switch (span->special.which) {
        case all_of_time: 
           result->special.which = negative_all_of_time; break;
        case negative_all_of_time: 
           result->special.which = all_of_time; break;
        case undefined:
           result->special.which = undefined; break;
        default: 
           internal_error("%s", "tadt_neg_fs: Unknown special span type");
      }
 
    default:
      internal_error("%s", "tadt_neg_fs: bad ts_type");

    }

  return ok; 
}

/*  
*  Routine:  tadt_fs_add_fs
*
*  Description: Add two fixed spans
*
*  Arguments: 	span1, span2 -- (IN) : spans to be added
*               result -- (OUT) : result span
*
*  Return Value: Error code
*
*  Errors: out_of_range
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_fs_add_fs(span_type span1, span_type span2, span_type result) {
  resolution_type flags;

  flags = combined_flags[span1->upack.ts_type][span2->upack.ts_type];
  switch (flags) {

    case upack32upack32: 
      ADD32(span1->upack32.granules,
            span2->upack32.granules,
            result->upack32.granules);
      return rc;

    case upack32upack64:
      LOAD64(span1->upack32.granules,temp64);
      ADD64(temp64, 
            span2->upack64.granules, 
            result->upack64.granules);
      return rc;

    case upack32upack96:
      LOAD96(span1->upack32.granules,temp96);
      ADD96(temp96, 
            span2->upack96.granules, 
            result->upack96.granules);
      return rc;

    case upack64upack32: 
      LOAD64(span2->upack32.granules,temp64);
      ADD64(span1->upack64.granules,
            temp64,
            result->upack64.granules);
      return rc;

    case upack64upack64:
      ADD64(span1->upack64.granules,
            span2->upack64.granules, 
            result->upack64.granules);
      return rc;

    case upack64upack96:
      LOAD96_FROM_64(span1->upack64.granules,temp96);
      ADD96(temp96, 
            span2->upack96.granules, 
            result->upack96.granules);
      return rc;

    case upack96upack32: 
      LOAD96(span2->upack32.granules,temp96);
      ADD96(span1->upack96.granules,
            temp96,
            result->upack96.granules);
      return rc;

    case upack96upack64:
      LOAD96_FROM_64(span2->upack64.granules,temp96);
      ADD96(span2->upack64.granules, 
            temp96,
            result->upack96.granules);
      return rc;

    case upack96upack96:
      ADD96(span1->upack96.granules, 
            span2->upack96.granules, 
            result->upack96.granules);
      return rc;

#ifdef INDETERMINATE
    case upack32upackind:
    case upack64upackind:
    case upack96upackind:
      result->upack.ts_type = upackind;
      result->upack.modified = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      rc = tadt_fs_add_fs(span1,
                          (span_type) &span2->upack_ind.lower,
                          (span_type) &result->upack_ind.lower);
      if (!rc) return rc;
      rc = tadt_fs_add_fs(span1,
                          (span_type) &span2->upack_ind.upper,
                          (span_type) &result->upack_ind.upper);
      return rc;

    case upackindupack32:
    case upackindupack64:
    case upackindupack96:
      result->upack.ts_type = upackind;
      result->upack.modified = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      rc = tadt_fs_add_fs((span_type) &span1->upack_ind.lower,
                          span2,
                          (span_type) &result->upack_ind.lower);
      if (!rc) return rc;
      rc = tadt_fs_add_fs((span_type) &span1->upack_ind.upper,
                          span2,
                          (span_type) &result->upack_ind.upper);
      return rc;

    case upackindupackind:
      result->upack.ts_type = upackind;
      result->upack.modified = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      rc = tadt_fs_add_fs((span_type) &span1->upack_ind.lower,
                          (span_type) &span2->upack_ind.lower,
                          (span_type) &result->upack_ind.lower);
      if (!rc) return rc;
      rc = tadt_fs_add_fs((span_type) &span1->upack_ind.upper,
                          (span_type) &span2->upack_ind.upper,
                          (span_type) &result->upack_ind.upper);
      return rc;
#endif 


#ifdef INDETERMINATE
    case upackindspecial:
#endif
    case upack32special:
    case upack64special:
    case upack96special:
      tadt_copy_s(span2, result);
      return ok;

#ifdef INDETERMINATE
    case specialupackind:
#endif
    case specialupack32: 
    case specialupack64: 
    case specialupack96: 
      tadt_copy_s(span1, result);
      return ok;

    case specialspecial: 
      result->special.ts_type = special;
      switch (span1->special.which) {
        case all_of_time: 
          switch (span2->special.which) {
            case all_of_time: 
              result->special.which = all_of_time; return ok;
            case negative_all_of_time: 
	      result->upack.ts_type = upack32; 
              result->upack.modified = TRUE;
              LOAD32(0,result->upack32.granules); 
	      return ok;
            case undefined:
              result->special.which = undefined; return ok;
            default: internal_error("%s", "tadt_fs_add_fs: Bad special type");
            }
        case negative_all_of_time: 
          switch (span2->special.which) {
            case all_of_time: 
	      result->upack.ts_type = upack32; 
              result->upack.modified = TRUE;
              result->upack.modified = TRUE;
              LOAD32(0,result->upack32.granules); 
	      return ok;
            case negative_all_of_time: 
              result->special.which = negative_all_of_time; return ok;
            case undefined:
              result->special.which = undefined; return ok;
            default: internal_error("%s", "tadt_fs_add_fs: Bad special type");
            }
        case undefined: 
          result->special.which = undefined; return ok;
        default: internal_error("%s", "tadt_fs_add_fs: Bad special type");
        }

    default:
      internal_error("%s", "tadt_fs_add_fs: BAD Flags");

  }

  internal_error("%s", "tadt_fs_add_fs: how did I get here?");
  return ok; /* should never get here */
}


/*  
*  Routine:  tadt_fs_minus_fs
*
*  Description: Subtract two fixed spans
*
*  Arguments: 	span1 -- (IN) : minuend (span to be subtracted from)
*		span2 -- (IN) : subtrahend
*               result -- (OUT) : result span
*
*  Return Value: Error code
*
*  Errors: out_of_range
*          One note: a loss of resolution is flagged as out of range.
*           (e.g., a hires + lowres --> hires or out_of_range)
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_fs_minus_fs(span_type span1, span_type span2, span_type result){
  resolution_type flags;

  flags = combined_flags[span1->upack.ts_type][span2->upack.ts_type];
  switch (flags) {
    case upack32upack32: 
      SUB32(span1->upack32.granules,
            span2->upack32.granules,
            result->upack32.granules);
      return rc;

    case upack32upack64:
      LOAD64(span1->upack32.granules,temp64);
      SUB64(temp64, 
            span2->upack64.granules, 
            result->upack64.granules);
      return rc;

    case upack32upack96:
      LOAD96(span1->upack32.granules,temp96);
      SUB96(temp96, 
            span2->upack96.granules, 
            result->upack96.granules);
      return rc;

    case upack64upack32: 
      LOAD64(span2->upack32.granules,temp64);
      SUB64(span1->upack64.granules,
            temp64,
            result->upack64.granules);
      return rc;

    case upack64upack64:
      SUB64(span1->upack64.granules,
            span2->upack64.granules, 
            result->upack64.granules);
      return rc;

    case upack64upack96:
      LOAD96_FROM_64(span1->upack64.granules,temp96);
      SUB96(temp96, 
            span2->upack96.granules, 
            result->upack96.granules);
      return rc;

    case upack96upack32: 
      LOAD96(span2->upack32.granules,temp96);
      SUB96(span1->upack96.granules,
            temp96,
            result->upack96.granules);
      return rc;

    case upack96upack64:
      LOAD96_FROM_64(span2->upack64.granules,temp96);
      SUB96(span2->upack64.granules, 
            temp96,
            result->upack96.granules);
      return rc;

    case upack96upack96:
      SUB96(span1->upack96.granules, 
            span2->upack96.granules, 
            result->upack96.granules);
      return rc;

#ifdef INDETERMINATE
    case upack32upackind:
    case upack64upackind:
    case upack96upackind:
      result->upack.ts_type = upackind;
      result->upack.modified = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      rc = tadt_fs_minus_fs(span1,
                          (span_type) &span2->upack_ind.lower,
                          (span_type) &result->upack_ind.lower);
      if (!rc) return rc;
      rc = tadt_fs_minus_fs(span1,
                          (span_type) &span2->upack_ind.upper,
                          (span_type) &result->upack_ind.upper);
      return rc;

    case upackindupack32:
    case upackindupack64:
    case upackindupack96:
      result->upack.ts_type = upackind;
      result->upack.modified = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      rc = tadt_fs_minus_fs((span_type) &span1->upack_ind.lower,
                          span2,
                          (span_type) &result->upack_ind.lower);
      if (!rc) return rc;
      rc = tadt_fs_minus_fs((span_type) &span1->upack_ind.upper,
                          span2,
                          (span_type) &result->upack_ind.upper);
      return rc;

    case upackindupackind:
      result->upack.ts_type = upackind;
      result->upack.modified = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      rc = tadt_fs_minus_fs((span_type) &span1->upack_ind.lower,
                          (span_type) &span2->upack_ind.lower,
                          (span_type) &result->upack_ind.lower);
      if (!rc) return rc;
      rc = tadt_fs_minus_fs((span_type) &span1->upack_ind.upper,
                          (span_type) &span2->upack_ind.upper,
                          (span_type) &result->upack_ind.upper);
      return rc;
#endif 

#ifdef INDETERMINATE
    case upackindspecial:
#endif
    case upack32special:
    case upack64special:
    case upack96special:
      result->special.ts_type = special;
      switch (span2->special.which) {
        case all_of_time: 
	  result->special.which = negative_all_of_time; return ok;
        case negative_all_of_time: 
	  result->special.which = all_of_time; return ok;
        case undefined: 
	  result->special.which = undefined; return ok;
        default: internal_error("%s", "tadt_fs_minus_fs: Bad special type");
	}

#ifdef INDETERMINATE
    case specialupackind:
#endif
    case specialupack32: 
    case specialupack64: 
    case specialupack96: 
      tadt_copy_s(span1, result);
      return ok;

    case specialspecial:
      result->special.ts_type = special;
      switch (span1->special.which) {
        case all_of_time: 
          switch (span2->special.which) {
            case all_of_time: 
	      result->upack.ts_type = upack32; 
              result->upack.modified = TRUE;
              LOAD32(0,result->upack32.granules); 
	      return ok;
            case negative_all_of_time: 
              result->special.which = all_of_time; return ok;
            case undefined:
              result->special.which = undefined; return ok;
            default: internal_error("%s", "tadt_fs_minus_fs: Bad special type");
            }
        case negative_all_of_time: 
          switch (span2->special.which) {
            case all_of_time: 
              result->special.which = negative_all_of_time; return ok;
            case negative_all_of_time: 
	      result->upack.ts_type = upack32; 
              result->upack.modified = TRUE;
              LOAD32(0,result->upack32.granules); 
	      return ok;
            case undefined:
              result->special.which = undefined; return ok;
            default: internal_error("%s", "tadt_fs_minus_fs: Bad special type");
            }
        case undefined: 
          result->special.which = undefined; return ok;
        default: internal_error("%s", "tadt_fs_minus_fs: Bad special type");
        }

    default:
      internal_error("%s", "tadt_fs_minus_fs: BAD Flags");

  }

  internal_error("%s", "tadt_fs_minus_fs: how did I get here?");
  return ok; /* should never get here */
}


/*  
*  Routine:  tadt_e_add_fs
*
*  Description: Displace an event by adding a span.  Result is always
*               an extended resolution timestamp except for adding a
*               lowres to lowres.  
*
*  Arguments: event -- (IN) : the event time
*	      span -- (IN) : the displacement
*             result -- (OUT) : result event
*
*  Return Value: Error code
*
*  Errors: out_of_range
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_e_add_fs(event_type event, span_type span, event_type result) {
  resolution_type flags;

  flags = combined_flags[event->upack.ts_type][span->upack.ts_type];
  switch (flags) {
    case upack32upack32: 
      ADD32(event->upack32.granules,
            span->upack32.granules,
            result->upack32.granules);
      return rc;

    case upack32upack64:
      LOAD64(event->upack32.granules,temp64);
      ADD64(temp64, 
            span->upack64.granules, 
            result->upack64.granules);
      return rc;

    case upack32upack96:
      LOAD96(event->upack32.granules,temp96);
      ADD96(temp96, 
            span->upack96.granules, 
            result->upack96.granules);
      return rc;

    case upack64upack32: 
      LOAD64(span->upack32.granules,temp64);
      ADD64(event->upack64.granules,
            temp64,
            result->upack64.granules);
      return rc;

    case upack64upack64:
      ADD64(event->upack64.granules,
            span->upack64.granules, 
            result->upack64.granules);
      return rc;

    case upack64upack96:
      LOAD96_FROM_64(event->upack64.granules,temp96);
      ADD96(temp96, 
            span->upack96.granules, 
            result->upack96.granules);
      return rc;

    case upack96upack32: 
      LOAD96(span->upack32.granules,temp96);
      ADD96(event->upack96.granules,
            temp96,
            result->upack96.granules);
      return rc;

    case upack96upack64:
      LOAD96_FROM_64(span->upack64.granules,temp96);
      ADD96(span->upack64.granules, 
            temp96,
            result->upack96.granules);
      return rc;

    case upack96upack96:
      ADD96(event->upack96.granules, 
            span->upack96.granules, 
            result->upack96.granules);
      return rc;

#ifdef INDETERMINATE
    case upack32upackind:
    case upack64upackind:
    case upack96upackind:
      result->upack.ts_type = upackind;
      result->upack.modified = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      rc = tadt_e_add_fs(event,
                          (span_type) &span->upack_ind.lower,
                          (event_type) &result->upack_ind.lower);
      if (!rc) return rc;
      rc = tadt_e_add_fs(event,
                          (span_type) &span->upack_ind.upper,
                          (event_type) &result->upack_ind.upper);
      return rc;

    case upackindupack32:
    case upackindupack64:
    case upackindupack96:
      result->upack.ts_type = upackind;
      result->upack.modified = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      rc = tadt_e_add_fs((event_type) &event->upack_ind.lower,
                          span,
                          (event_type) &result->upack_ind.lower);
      if (!rc) return rc;
      rc = tadt_e_add_fs((event_type) &event->upack_ind.upper,
                          span,
                          (event_type) &result->upack_ind.upper);
      return rc;

    case upackindupackind:
      result->upack.ts_type = upackind;
      result->upack.modified = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      rc = tadt_e_add_fs((event_type) &event->upack_ind.lower,
                          (span_type) &span->upack_ind.lower,
                          (event_type) &result->upack_ind.lower);
      if (!rc) return rc;
      rc = tadt_e_add_fs((event_type) &event->upack_ind.upper,
                          (span_type) &span->upack_ind.upper,
                          (event_type) &result->upack_ind.upper);
      return rc;
#endif 

#ifdef INDETERMINATE
    case upackindspecial:
#endif
    case upack32special:
    case upack64special:
    case upack96special:
      switch (span->special.which) {
	case all_of_time: 
           result->special.which = forever;
           result->special.ts_type = special;
           return ok;

	case negative_all_of_time:  
           result->special.which = beginning;
           result->special.ts_type = special;
           return ok;

	case undefined: 
           result->special.which = undefined;
           result->special.ts_type = special;
           return ok;

        default:
           internal_error("%s", "tadt_e_add_fs: bad special type");
	}

#ifdef INDETERMINATE
    case specialupackind:
#endif
    case specialupack32: 
    case specialupack64: 
    case specialupack96: 
      tadt_copy_e(event, result);
      return ok;

    case specialspecial: /* special and special */
      result->special.ts_type = special;
      switch (event->special.which) {
        case beginning: 
        case forever: 
          switch (span->special.which) {
            case all_of_time: 
              result->special.which = forever; return ok;
            case negative_all_of_time: 
              result->special.which = beginning; return ok;
            case undefined:
              result->special.which = undefined; return ok;
            default: internal_error("%s", "tadt_e_add_fs: Bad special type");
            }
        case undefined: 
          result->special.which = undefined; return ok;
        default: internal_error("%s", "tadt_e_add_fs: Bad special type");
        }

    default:
      internal_error("%s %x", "tadt_e_add_fs: BAD Flags, must unpack event/span", flags);

  }

  internal_error("%s", "tadt_e_add_fs: how did I get here?");
  return ok; /* should never get here */
}

/*  
*  Routine:  tadt_e_minus_fs
*
*  Description: Displace an event by subtracting a span 
*
*  Arguments: event -- (IN) : the event time
*	      span  -- (IN) : the displacement
*             result -- (OUT) : result event
*
*  Return Value: Error code
*
*  Errors: out_of_range
*          overflow  (beginning - anything)
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_e_minus_fs(event_type event, span_type span, event_type result) {
  resolution_type flags;

  flags = combined_flags[event->upack.ts_type][span->upack.ts_type];
  switch (flags) {

    case upack32upack32: 
      SUB32(event->upack32.granules,
            span->upack32.granules,
            result->upack32.granules);
      return rc;

    case upack32upack64:
      LOAD64(event->upack32.granules,temp64);
      SUB64(temp64, 
            span->upack64.granules, 
            result->upack64.granules);
      return rc;

    case upack32upack96:
      LOAD96(event->upack32.granules,temp96);
      SUB96(temp96, 
            span->upack96.granules, 
            result->upack96.granules);
      return rc;

    case upack64upack32: 
      LOAD64(span->upack32.granules,temp64);
      SUB64(event->upack64.granules,
            temp64,
            result->upack64.granules);
      return rc;

    case upack64upack64:
      SUB64(event->upack64.granules,
            span->upack64.granules, 
            result->upack64.granules);
      return rc;

    case upack64upack96:
      LOAD96_FROM_64(event->upack64.granules,temp96);
      SUB96(temp96, 
            span->upack96.granules, 
            result->upack96.granules);
      return rc;

    case upack96upack32: 
      LOAD96(span->upack32.granules,temp96);
      SUB96(event->upack96.granules,
            temp96,
            result->upack96.granules);
      return rc;

    case upack96upack64:
      LOAD96_FROM_64(span->upack64.granules,temp96);
      SUB96(span->upack64.granules, 
            temp96,
            result->upack96.granules);
      return rc;

    case upack96upack96:
      SUB96(event->upack96.granules, 
            span->upack96.granules, 
            result->upack96.granules);
      return rc;

#ifdef INDETERMINATE
    case upack32upackind:
    case upack64upackind:
    case upack96upackind:
      result->upack.ts_type = upackind;
      result->upack.modified = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      rc = tadt_e_minus_fs(event,
                          (span_type) &span->upack_ind.lower,
                          (event_type) &result->upack_ind.lower);
      if (!rc) return rc;
      rc = tadt_e_minus_fs(event,
                          (span_type) &span->upack_ind.upper,
                          (event_type) &result->upack_ind.upper);
      return rc;

    case upackindupack32:
    case upackindupack64:
    case upackindupack96:
      result->upack.ts_type = upackind;
      result->upack.modified = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      rc = tadt_e_minus_fs((event_type) &event->upack_ind.lower,
                          span,
                          (event_type) &result->upack_ind.lower);
      if (!rc) return rc;
      rc = tadt_e_minus_fs((event_type) &event->upack_ind.upper,
                          span,
                          (event_type) &result->upack_ind.upper);
      return rc;

    case upackindupackind:
      result->upack.ts_type = upackind;
      result->upack.modified = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      rc = tadt_e_minus_fs((event_type) &event->upack_ind.lower,
                          (span_type) &span->upack_ind.lower,
                          (event_type) &result->upack_ind.lower);
      if (!rc) return rc;
      rc = tadt_e_minus_fs((event_type) &event->upack_ind.upper,
                          (span_type) &span->upack_ind.upper,
                          (event_type) &result->upack_ind.upper);
      return rc;
#endif 

#ifdef INDETERMINATE
    case upackindspecial:
#endif
    case upack32special:
    case upack64special:
    case upack96special:
      switch (span->special.which) {
	case all_of_time: 
           result->special.which = beginning;
           result->special.ts_type = special;
           return ok;

	case negative_all_of_time:  
           result->special.which = forever;
           result->special.ts_type = special;
           return ok;

	case undefined: 
           result->special.which = undefined;
           result->special.ts_type = special;
           return ok;

        default:
           internal_error("%s", "tadt_e_minus_fs: bad special type");
	}

#ifdef INDETERMINATE
    case specialupackind:
#endif
    case specialupack32: 
    case specialupack64: 
    case specialupack96: 
      tadt_copy_e(event, result);
      return ok;

    case specialspecial: /* special and special */
      result->special.ts_type = special;
      switch (event->special.which) {
        case beginning: 
        case forever: 
          switch (span->special.which) {
            case all_of_time: 
              result->special.which = beginning; return ok;
            case negative_all_of_time: 
              result->special.which = forever; return ok;
            case undefined:
              result->special.which = undefined; return ok;
            default: internal_error("%s", "tadt_e_minus_fs: Bad special type");
            }
        case undefined: 
          result->special.which = undefined; return ok;
        default: internal_error("%s", "tadt_e_minus_fs: Bad special type");
        }

    default:
      internal_error("%s", "tadt_e_minus_fs: BAD Flags");

  }

  internal_error("%s", "tadt_e_minus_fs: how did I get here?");
  return ok; /* should never get here */
}

/*  
*  Routine:  tadt_e_minus_e
*
*  Description: Compute the span between two events by 
*		subtracting the second event from the first event (e1 - e2).
*		if e1 > e2,  return a positive span
*		if e1 < e2,  return a negative span
*
*  Arguments:   event1, event2 -- (IN) : two input events.
*               result -- (OUT) : result span
*
*  Return Value: Error code
*
*  Errors: none
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_e_minus_e(event_type event1, event_type event2, span_type result) {
  resolution_type flags;

  flags = combined_flags[event1->upack.ts_type][event2->upack.ts_type];
  switch (flags) {

    case upack32upack32: 
      SUB32(event1->upack32.granules,
            event2->upack32.granules,
            result->upack32.granules);
      return rc;

    case upack32upack64:
      LOAD64(event1->upack32.granules,temp64);
      SUB64(temp64, 
            event2->upack64.granules, 
            result->upack64.granules);
      return rc;

    case upack32upack96:
      LOAD96(event1->upack32.granules,temp96);
      SUB96(temp96, 
            event2->upack96.granules, 
            result->upack96.granules);
      return rc;

    case upack64upack32: 
      LOAD64(event2->upack32.granules,temp64);
      SUB64(event1->upack64.granules,
            temp64,
            result->upack64.granules);
      return rc;

    case upack64upack64:
      SUB64(event1->upack64.granules,
            event2->upack64.granules, 
            result->upack64.granules);
      return rc;

    case upack64upack96:
      LOAD96_FROM_64(event1->upack64.granules,temp96);
      SUB96(temp96, 
            event2->upack96.granules, 
            result->upack96.granules);
      return rc;

    case upack96upack32: 
      LOAD96(event2->upack32.granules,temp96);
      SUB96(event1->upack96.granules,
            temp96,
            result->upack96.granules);
      return rc;

    case upack96upack64:
      LOAD96_FROM_64(event2->upack64.granules,temp96);
      SUB96(event2->upack64.granules, 
            temp96,
            result->upack96.granules);
      return rc;

    case upack96upack96:
      SUB96(event1->upack96.granules, 
            event2->upack96.granules, 
            result->upack96.granules);
      return rc;

#ifdef INDETERMINATE
    case upack32upackind:
    case upack64upackind:
    case upack96upackind:
      result->upack.ts_type = upackind;
      result->upack.modified = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      rc = tadt_e_minus_e(event1,
                          (event_type) &event2->upack_ind.lower,
                          (span_type) &result->upack_ind.lower);
      if (!rc) return rc;
      rc = tadt_e_minus_e(event1,
                          (event_type) &event2->upack_ind.upper,
                          (span_type) &result->upack_ind.upper);
      return rc;

    case upackindupack32:
    case upackindupack64:
    case upackindupack96:
      result->upack.ts_type = upackind;
      result->upack.modified = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      rc = tadt_e_minus_e((event_type) &event1->upack_ind.lower,
                          event2,
                          (span_type) &result->upack_ind.lower);
      if (!rc) return rc;
      rc = tadt_e_minus_e((event_type) &event1->upack_ind.upper,
                          event2,
                          (span_type) &result->upack_ind.upper);
      return rc;

    case upackindupackind:
      result->upack.ts_type = upackind;
      result->upack.modified = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      rc = tadt_e_minus_e((event_type) &event1->upack_ind.lower,
                          (event_type) &event2->upack_ind.lower,
                          (span_type) &result->upack_ind.lower);
      if (!rc) return rc;
      rc = tadt_e_minus_e((event_type) &event1->upack_ind.upper,
                          (event_type) &event2->upack_ind.upper,
                          (span_type) &result->upack_ind.upper);
      return rc;
#endif 

#ifdef INDETERMINATE
    case upackindspecial:
#endif
    case upack32special:
    case upack64special:
    case upack96special:
      switch (event2->special.which) {
	case forever: 
           result->special.which = negative_all_of_time;
           result->upack.ts_type = special;
           return ok;

	case beginning:  
           result->special.which = all_of_time;
           result->upack.ts_type = special;
           return ok;

	case undefined: 
           result->special.which = undefined;
           result->upack.ts_type = special;
           return ok;

        default:
           internal_error("%s", "tadt_e_minus_e: bad special type");
	}


#ifdef INDETERMINATE
    case specialupackind:
#endif
    case specialupack32:
    case specialupack64:
    case specialupack96:
      switch (event1->special.which) {
	case forever: 
           result->special.which = all_of_time;
           result->special.ts_type = special;
           return ok;

	case beginning:  
           result->special.which = negative_all_of_time;
           result->special.ts_type = special;
           return ok;

	case undefined: 
           result->special.which = undefined;
           result->special.ts_type = special;
           return ok;

        default:
           internal_error("%s", "tadt_e_minus_e: bad special type");
	}

    case specialspecial: /* special and special */
      result->upack.ts_type = special;
      switch (event1->special.which) {
        case beginning: 
          switch (event2->special.which) {
            case beginning: 
	      result->upack.ts_type = upack; 
	      result->upack.modified = TRUE; 
	      result->upack32.granules = 0;
	      return ok;
            case forever: 
              result->special.which = negative_all_of_time; return ok;
            case undefined:
              result->special.which = undefined; return ok;
            default: internal_error("%s", "tadt_e_minus_e: Bad special type");
            }
        case forever: 
          switch (event2->special.which) {
            case beginning: 
              result->special.which = all_of_time; return ok;
            case forever: 
	      result->upack.ts_type = upack; 
	      result->upack.modified = TRUE; 
	      result->upack32.granules = 0;
	      return ok;
            case undefined:
              result->special.which = undefined; return ok;
            default: internal_error("%s", "tadt_e_minus_e: Bad special type");
            }

        case undefined: 
          result->special.which = undefined; return ok;
        default: internal_error("%s", "tadt_e_minus_e: Bad special type");
        }

    default:
      internal_error("%s", "tadt_e_minus_e: BAD Flags");

  }

  internal_error("%s", "tadt_e_minus_e: how did I get here?");
  return ok; /* should never get here */
}

/*  
*  Routine:  tadt_fs_times_n
*
*  Description: Return a span equal to a multiple of the input span.  The
*               multiplier must be an integer.
*
*  Arguments:   span -- (IN): the input span
*		n -- (IN): number of times
*               result -- (OUT) : result span
*
*  Return Value: Error code
*
*  Errors: out_of_range
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_fs_times_n(span_type span, int n, span_type result) {

  result->upack.modified = TRUE;
  switch (span->upack.ts_type) {
    case upack32: 
      result->upack.ts_type = upack32;
      LOAD32(n,temp32);
      MULT32(temp32,span->upack32.granules,result->upack32.granules); 
      return rc;

    case upack64: 
      result->upack.ts_type = upack64;
      LOAD64(n,temp64); 
      MULT64(temp64,span->upack64.granules,result->upack64.granules); 
      return rc;

    case upack96: 
      result->upack.ts_type = upack96;
      LOAD96(n,temp96); 
      MULT96(temp96,span->upack96.granules,result->upack96.granules); 
      return rc;

#ifdef INDETERMINATE
    case upackind:
      result->upack.ts_type = upackind;
      rc = tadt_fs_times_n((span_type) &span->upack_ind.lower,
                           n,
                           (span_type) &result->upack_ind.lower);
      if (!rc) return rc;
      rc = tadt_fs_times_n((span_type) &span->upack_ind.upper,
                           n,
                           (span_type) &result->upack_ind.upper);
      return rc;
#endif

    case special:
      result->special.ts_type = special;
      switch (span->special.which) {
        case all_of_time:
          result->special.which = (n < 0)? negative_all_of_time: all_of_time;
          break;
        case negative_all_of_time:
          result->special.which = (n < 0)? all_of_time: negative_all_of_time;
          break;
        case undefined:
          result->special.which = undefined;
          break;
        default: internal_error("%s", "tadt_fs_times_n: Bad special type");
        }
    default: internal_error("%s", "tadt_fs_times_n: Bad ts_type");
    }
  return ok;
}

/*  
*  Routine:  tadt_fs_div_n
*
*  Description: Divide a span by a numeric value.  N must be nonzero.
*
*  Arguments: span -- (IN):the input span
*	      n -- (IN): number of division
*             result -- (OUT) : result span
*
*  Return Value: Error code
*
*  Errors: out_of_range
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_fs_div_n(span_type span, int n, span_type result) {

  /* Special case division by zero */
  if (n == 0) {
    tadt_copy_s(span, result);
    return divide_by_zero;
    }

  result->upack.modified = TRUE;
  switch (span->upack.ts_type) {

    case upack32: 
      result->upack.ts_type = upack32;
      LOAD32(n,temp32);
      DIV32(span->upack32.granules,temp32,result->upack32.granules); 
      return rc;

    case upack64: 
      result->upack.ts_type = upack64;
      LOAD64(n,temp64); 
      DIV64(span->upack64.granules,temp64,result->upack64.granules); 
      return rc;

    case upack96: 
      result->upack.ts_type = upack96;
      LOAD96(n,temp96); 
      DIV96(span->upack96.granules,temp96,result->upack96.granules); 
      return rc;

#ifdef INDETERMINATE
    case upackind:
      result->upack.ts_type = upackind;
      rc = tadt_fs_div_n((span_type) &span->upack_ind.lower,
                           n,
                           (span_type) &result->upack_ind.lower);
      if (!rc) return rc;
      rc = tadt_fs_div_n((span_type) &span->upack_ind.upper,
                           n,
                           (span_type) &result->upack_ind.upper);
      return rc;
#endif

    case special:
      result->special.ts_type = special;
      switch (span->special.which) {
        case all_of_time:
          result->special.which = (n < 0)? negative_all_of_time: all_of_time;
          break;
        case negative_all_of_time:
          result->special.which = (n < 0)? all_of_time: negative_all_of_time;
          break;
        case undefined:
          result->special.which = undefined;
          break;
        default: internal_error("%s", "tadt_fs_div_n: Bad special type");
        }
    default: internal_error("%s", "tadt_fs_div_n: Bad ts_type");
    }
  return ok;
}

/*  
*  Routine:  tadt_fs_div_fs
*
*  Description: Divide one span by another.  
*
*  Arguments:   span1 -- (IN): the dividend
*	  	span2 --(IN): the divisor
*               result -- (OUT) : result integer (double)
*
*  Return Value: Error code
*
*  Errors: division_by_zero
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_fs_div_fs(span_type span1, span_type span2, double *result){
  resolution_type flags;

  flags = combined_flags[span1->upack.ts_type][span2->upack.ts_type];
  switch (flags) {

    case upack32upack32: 
      LOAD32(0,temp32);
      if (EQ32(temp32,span2->upack32.granules)) {
        *result = 0;
        return divide_by_zero;
        }
      FDIV32(span1->upack32.granules,
            span2->upack32.granules,
            *result);
      return rc;

    case upack32upack64:
      LOAD64(0,temp64);
      if (EQ64(temp64,span2->upack64.granules)) {
        *result = 0;
        return divide_by_zero;
        }
      LOAD64(span1->upack32.granules,temp64);
      FDIV64(temp64, 
            span2->upack64.granules, 
            *result);
      return rc;

    case upack32upack96:
      LOAD96(0,temp96);
      if (EQ96(temp96,span2->upack96.granules)) {
        *result = 0;
        return divide_by_zero;
        }
      LOAD96(span1->upack32.granules,temp96);
      FDIV96(temp96, 
            span2->upack96.granules, 
            *result);
      return rc;

    case upack64upack32: 
      LOAD32(0,temp32);
      if (EQ32(temp32,span2->upack32.granules)) {
        *result = 0;
        return divide_by_zero;
        }
      LOAD64(span2->upack32.granules,temp64);
      FDIV64(span1->upack64.granules,
            temp64,
            *result);
      return rc;

    case upack64upack64:
      LOAD64(0,temp64);
      if (EQ64(temp64,span2->upack64.granules)) {
        *result = 0;
        return divide_by_zero;
        }
      FDIV64(span1->upack64.granules,
            span2->upack64.granules, 
            *result);
      return rc;

    case upack64upack96:
      LOAD96(0,temp96);
      if (EQ96(temp96,span2->upack96.granules)) {
        *result = 0;
        return divide_by_zero;
        }
      LOAD96_FROM_64(span1->upack64.granules,temp96);
      FDIV96(temp96, 
            span2->upack96.granules, 
            *result);
      return rc;

    case upack96upack32: 
      LOAD32(0,temp32);
      if (EQ32(temp32,span2->upack32.granules)) {
        *result = 0;
        return divide_by_zero;
        }
      LOAD96(span2->upack32.granules,temp96);
      FDIV96(span1->upack96.granules,
            temp96,
            *result);
      return rc;

    case upack96upack64:
      LOAD64(0,temp64);
      if (EQ64(temp64,span2->upack64.granules)) {
        *result = 0;
        return divide_by_zero;
        }
      LOAD96_FROM_64(span2->upack64.granules,temp96);
      FDIV96(span2->upack64.granules, 
            temp96,
            *result);
      return rc;

    case upack96upack96:
      LOAD96(0,temp96);
      if (EQ96(temp96,span2->upack96.granules)) {
        *result = 0;
        return divide_by_zero;
        }
      FDIV96(span1->upack96.granules, 
            span2->upack96.granules, 
            *result);
      return rc;

#ifdef INDETERMINATE
    case upack32upackind:
    case upack64upackind:
    case upack96upackind:
      rc = tadt_fs_div_fs(span1,
                          (span_type) &span2->upack_ind.lower,
                          result);
      if (!rc) return rc;
      rc = tadt_fs_div_fs(span1,
                          (span_type) &span2->upack_ind.upper,
                          result);
      return rc;

    case upackindupack32:
    case upackindupack64:
    case upackindupack96:
      rc = tadt_fs_div_fs((span_type) &span1->upack_ind.lower,
                          span2,
                          result);
      if (!rc) return rc;
      rc = tadt_fs_div_fs((span_type) &span1->upack_ind.upper,
                          span2,
                          result);
      return rc;

    case upackindupackind:
      rc = tadt_fs_div_fs((span_type) &span1->upack_ind.lower,
                          (span_type) &span2->upack_ind.lower,
                          result);
      if (!rc) return rc;
      rc = tadt_fs_div_fs((span_type) &span1->upack_ind.upper,
                          (span_type) &span2->upack_ind.upper,
                          result);
      return rc;
#endif 

#ifdef INDETERMINATE
    case upackindspecial:
#endif
    case upack32special:
    case upack64special:
    case upack96special:
      *result = 0;
      return ok;

#ifdef INDETERMINATE
    case specialupackind:
#endif
    case specialupack32:
    case specialupack64:
    case specialupack96:
      return out_of_range;

    case specialspecial: 
      switch (span1->special.which) {
        case all_of_time: 
          switch (span2->special.which) {
            case all_of_time: *result = 1; return ok;
            case negative_all_of_time: *result = -1; return ok;
            default: internal_error("%s", "tadt_fs_div_fs: Bad special type");
            }
        case negative_all_of_time: 
          switch (span2->special.which) {
            case all_of_time: *result = -1; return ok;
            case negative_all_of_time: *result = 1; return ok;
            default: internal_error("%s", "tadt_fs_div_fs: Bad special type");
            }
        default: internal_error("%s", "tadt_fs_div_fs: Bad special type");
        }

    default:
      internal_error("%s", "tadt_fs_div_fs: BAD Flags");

  }

  internal_error("%s", "tadt_fs_div_fs: how did I get here?");
  return ok; /* should never get here */
}

/*  
*  Routine:  tadt_i_add_fs
*
*  Description: Displace an interval by adding a span
*
*  Arguments: interval -- (IN): the interval to be displaced
*	      span -- (IN): the span of displacement
*             result -- (OUT) : result interval
*
*  Return Value: Error code
*
*  Errors: out_of_range
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_i_add_fs(interval_type interval, span_type span, interval_type result) {
  error_type rc;

  /* Displace each bounding event */
  if ((rc = tadt_e_add_fs(interval->start, span, result->start)) != ok)
     return rc;
  return tadt_e_add_fs(interval->terminate, span, result->terminate);
}

/*  
*  Routine:  tadt_i_minus_fs
*
*  Description: Displace an interval by subtracting a span
*
*  Arguments: interval -- (IN): the interval to be displaced
*	      span -- (IN): the span of displacement
*             result -- (OUT) : result interval
*
*  Return Value: Error code
*
*  Errors: out_of_range
*          overflow     (|beginning,forever| - anything)
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_i_minus_fs(interval_type interval, span_type span, interval_type result) {
  error_type rc;

  /* Displace each endpoint separately */
  if ((rc = tadt_e_minus_fs(interval->start, span, result->start)) != ok) return rc;
  return tadt_e_minus_fs(interval->terminate, span, result->terminate);
}  
